import inspect
from concurrent.futures import Future


def call_func_with_kwds(func, args, kwds, log=None):
    funcname = getattr(func, '__name__', repr(func))
    sig = inspect.signature(func)
    for p in sig.parameters.values():
        if p.kind == p.VAR_KEYWORD:
            return func(*args, **kwds)
    kw = {
        k: v
        for k, v in kwds.items()
        if k in list(sig.parameters.keys())[len(args):]
    }
    try:
        args = [
            arg.result() if isinstance(arg, Future) else arg for arg in args
        ]
        kw = {
            k: v.result() if isinstance(v, Future) else v
            for k, v in kw.items()
        }
        return func(*args, **kw)
    except:
        if log:
            log.exception(f'Call {funcname} with {args} and {kw}')
        raise
    finally:
        if log:
            log.debug(f'Call {funcname} with {args} and {kw}')


def try_to_call(x, args, kwds, log=None):
    if callable(x):
        return call_func_with_kwds(x, args, kwds, log)
    return x



import asyncio
import pickle
import sys
import time
import uuid
from pathlib import Path

import click
import dill
import numpy as np
import zmq
from loguru import logger

from qulab.sys.rpc.zmq_socket import ZMQContextManager

from .curd import query_record, remove_tags, tag, update_tags
from .models import Record as RecordInDB
from .models import Session, create_engine, create_tables, sessionmaker, utcnow

_notgiven = object()
datapath = Path.home() / 'qulab' / 'data'
datapath.mkdir(parents=True, exist_ok=True)

record_cache = {}


def random_path(base):
    while True:
        s = uuid.uuid4().hex
        path = base / s[:2] / s[2:4] / s[4:6] / s[6:]
        if not path.exists():
            return path


class BufferList():

    def __init__(self, pos_file=None, value_file=None):
        self._pos = []
        self._value = []
        self.lu = ()
        self.rd = ()
        self.pos_file = pos_file
        self.value_file = value_file

    @property
    def shape(self):
        return tuple([i - j for i, j in zip(self.rd, self.lu)])

    def flush(self):
        if self.pos_file is not None:
            with open(self.pos_file, 'ab') as f:
                for pos in self._pos:
                    dill.dump(pos, f)
            self._pos.clear()
        if self.value_file is not None:
            with open(self.value_file, 'ab') as f:
                for value in self._value:
                    dill.dump(value, f)
            self._value.clear()

    def append(self, pos, value):
        self.lu = tuple([min(i, j) for i, j in zip(pos, self.lu)])
        self.rd = tuple([max(i + 1, j) for i, j in zip(pos, self.rd)])
        self._pos.append(pos)
        self._value.append(value)
        if len(self._value) > 1000:
            self.flush()

    def value(self):
        v = []
        if self.value_file is not None:
            with open(self.value_file, 'rb') as f:
                while True:
                    try:
                        v.append(dill.load(f))
                    except EOFError:
                        break
        v.extend(self._value)
        return v

    def pos(self):
        p = []
        if self.pos_file is not None:
            with open(self.pos_file, 'rb') as f:
                while True:
                    try:
                        p.append(dill.load(f))
                    except EOFError:
                        break
        p.extend(self._pos)
        return p

    def array(self):
        pos = np.asarray(self.pos()) - np.asarray(self.lu)
        data = np.asarray(self.value())
        inner_shape = data.shape[1:]
        x = np.full(self.shape + inner_shape, np.nan, dtype=data[0].dtype)
        x.__setitem__(tuple(pos.T), data)
        return x


class Record():

    def __init__(self, id, database, description=None):
        self.id = id
        self.database = database
        self.description = description
        self._keys = set()
        self._items = {}
        self._index = []
        self._pos = []
        self._last_vars = set()
        self._levels = {}
        self._file = None
        self.independent_variables = {}
        self.constants = {}

        for level, group in self.description['order'].items():
            for names in group:
                for name in names:
                    self._levels[name] = level

        for name, value in self.description['consts'].items():
            if name not in self._items:
                self._items[name] = value
            self.constants[name] = value
        for level, range_list in self.description['loops'].items():
            for name, iterable in range_list:
                if isinstance(iterable, (np.ndarray, list, tuple, range)):
                    self._items[name] = iterable
                    self.independent_variables[name] = (level, iterable)

        if self.is_local_record():
            self.database = Path(self.database)
            self._file = random_path(self.database / 'objects')
            self._file.parent.mkdir(parents=True, exist_ok=True)

    def is_local_record(self):
        return not self.is_cache_record() and not self.is_remote_record()

    def is_cache_record(self):
        return self.database is None

    def is_remote_record(self):
        return isinstance(self.database,
                          str) and self.database.startswith("tcp://")

    def __del__(self):
        self.flush()

    def __getitem__(self, key):
        return self.get(key)

    def get(self, key, default=_notgiven, buffer_to_array=True):
        if self.is_remote_record():
            with ZMQContextManager(zmq.DEALER,
                                   connect=self.database) as socket:
                socket.send_pyobj({
                    'method': 'record_getitem',
                    'record_id': self.id,
                    'key': key
                })
                ret = socket.recv_pyobj()
                if isinstance(ret, BufferList) and buffer_to_array:
                    return ret.array()
                else:
                    return ret
        else:
            if default is _notgiven:
                d = self._items.get(key)
            else:
                d = self._items.get(key, default)
            if isinstance(d, BufferList):
                if buffer_to_array:
                    return d.array()
                else:
                    ret = BufferList()
                    ret._pos = d.pos()
                    ret._value = d.value()
                    ret.lu = d.lu
                    ret.rd = d.rd
                    return ret
            else:
                return d

    def keys(self):
        if self.is_remote_record():
            with ZMQContextManager(zmq.DEALER,
                                   connect=self.database) as socket:
                socket.send_pyobj({
                    'method': 'record_keys',
                    'record_id': self.id
                })
                return socket.recv_pyobj()
        else:
            return list(self._keys)

    def append(self, level, step, position, variables):
        if level < 0:
            self.flush()
            return

        for key in set(variables.keys()) - self._last_vars:
            if key not in self._levels:
                self._levels[key] = level

        self._last_vars = set(variables.keys())
        self._keys.update(variables.keys())

        if level >= len(self._pos):
            l = level + 1 - len(self._pos)
            self._index.extend(([0] * (l - 1)) + [step])
            self._pos.extend(([0] * (l - 1)) + [position])
            pos = tuple(self._pos)
        elif level == len(self._pos) - 1:
            self._index[-1] = step
            self._pos[-1] = position
            pos = tuple(self._pos)
        else:
            self._index = self._index[:level + 1]
            self._pos = self._pos[:level + 1]
            self._index[-1] = step + 1
            self._pos[-1] = position
            pos = tuple(self._pos)
            self._pos[-1] += 1

        for key, value in variables.items():
            if level == self._levels[key]:
                if key not in self._items:
                    if self.is_local_record():
                        f1 = random_path(self.database / 'objects')
                        f1.parent.mkdir(parents=True, exist_ok=True)
                        f2 = random_path(self.database / 'objects')
                        f2.parent.mkdir(parents=True, exist_ok=True)
                        self._items[key] = BufferList(f1, f2)
                    else:
                        self._items[key] = BufferList()
                    self._items[key].lu = pos
                    self._items[key].rd = tuple([i + 1 for i in pos])
                    self._items[key].append(pos, value)
                elif isinstance(self._items[key], BufferList):
                    self._items[key].append(pos, value)
            elif self._levels[key] == -1 and key not in self._items:
                self._items[key] = value

    def flush(self):
        if self.is_remote_record() or self.is_cache_record():
            return

        for key, value in self._items.items():
            if isinstance(value, BufferList):
                value.flush()

        with open(self._file, 'wb') as f:
            dill.dump(self, f)


class Request():
    __slots__ = ['sock', 'identity', 'msg', 'method']

    def __init__(self, sock, identity, msg):
        self.sock = sock
        self.identity = identity
        self.msg = pickle.loads(msg)
        self.method = self.msg.get('method', '')


async def reply(req, resp):
    await req.sock.send_multipart([req.identity, pickle.dumps(resp)])


def clear_cache():
    if len(record_cache) < 1024:
        return

    for k, (t, _) in zip(sorted(record_cache.items(), key=lambda x: x[1][0]),
                         range(len(record_cache) - 1024)):
        del record_cache[k]


def flush_cache():
    for k, (t, r) in record_cache.items():
        r.flush()


def get_record(session, id, datapath):
    if id not in record_cache:
        record_in_db = session.get(RecordInDB, id)
        record_in_db.atime = utcnow()
        path = datapath / 'objects' / record_in_db.file
        with open(path, 'rb') as f:
            record = dill.load(f)
    else:
        record = record_cache[id][1]
    clear_cache()
    record_cache[id] = time.time(), record
    return record


def create_record(session, description, datapath):
    record = Record(None, datapath, description)
    record_in_db = RecordInDB()
    if 'app' in description:
        record_in_db.app = description['app']
    if 'tags' in description:
        record_in_db.tags = [tag(session, t) for t in description['tags']]
    record_in_db.file = '/'.join(record._file.parts[-4:])
    session.add(record_in_db)
    try:
        session.commit()
        record.id = record_in_db.id
        clear_cache()
        record_cache[record.id] = time.time(), record
        return record.id
    except:
        session.rollback()
        raise


@logger.catch
async def handle(session: Session, request: Request, datapath: Path):

    msg = request.msg

    match request.method:
        case 'ping':
            await reply(request, 'pong')
        case 'record_create':
            description = dill.loads(msg['description'])
            await reply(request, create_record(session, description, datapath))
        case 'record_append':
            record = get_record(session, msg['record_id'], datapath)
            record.append(msg['level'], msg['step'], msg['position'],
                          msg['variables'])
            if msg['level'] < 0:
                record_in_db = session.get(RecordInDB, msg['record_id'])
                record_in_db.mtime = utcnow()
                record_in_db.atime = utcnow()
                session.commit()
        case 'record_description':
            record = get_record(session, msg['record_id'], datapath)
            await reply(request, dill.dumps(record.description))
        case 'record_getitem':
            record = get_record(session, msg['record_id'], datapath)
            await reply(request, record.get(msg['key'], buffer_to_array=False))
        case 'record_keys':
            record = get_record(session, msg['record_id'], datapath)
            await reply(request, record.keys())
        case 'record_query':
            total, apps, table = query_record(session,
                                              offset=msg.get('offset', 0),
                                              limit=msg.get('limit', 10),
                                              app=msg.get('app', None),
                                              tags=msg.get('tags', ()),
                                              before=msg.get('before', None),
                                              after=msg.get('after', None))
            await reply(request, (total, apps, table))
        case 'record_get_tags':
            record_in_db = session.get(RecordInDB, msg['record_id'])
            await reply(request, [t.name for t in record_in_db.tags])
        case 'record_remove_tags':
            remove_tags(session, msg['record_id'], msg['tags'])
        case 'record_add_tags':
            update_tags(session, msg['record_id'], msg['tags'], True)
        case 'record_replace_tags':
            update_tags(session, msg['record_id'], msg['tags'], False)
        case _:
            logger.error(f'Unknown method: {msg["method"]}')


async def _handle(session: Session, request: Request, datapath: Path):
    try:
        await handle(session, request, datapath)
    except:
        await reply(request, 'error')


async def serv(port,
               datapath,
               url=None,
               buffer_size=1024 * 1024 * 1024,
               interval=60):
    logger.info('Server starting.')
    async with ZMQContextManager(zmq.ROUTER, bind=f"tcp://*:{port}") as sock:
        if url is None:
            url = 'sqlite:///' + str(datapath / 'data.db')
        engine = create_engine(url)
        create_tables(engine)
        Session = sessionmaker(engine)
        with Session() as session:
            logger.info('Server started.')
            received = 0
            last_flush_time = time.time()
            while True:
                identity, msg = await sock.recv_multipart()
                received += len(msg)
                req = Request(sock, identity, msg)
                asyncio.create_task(_handle(session, req, datapath))
                if received > buffer_size or time.time(
                ) - last_flush_time > interval:
                    flush_cache()
                    received = 0
                    last_flush_time = time.time()


async def watch(port, datapath, url=None, timeout=1, buffer=1024, interval=60):
    with ZMQContextManager(zmq.DEALER,
                           connect=f"tcp://127.0.0.1:{port}") as sock:
        sock.setsockopt(zmq.LINGER, 0)
        while True:
            try:
                sock.send_pyobj({"method": "ping"})
                if sock.poll(int(1000 * timeout)):
                    sock.recv()
                else:
                    raise asyncio.TimeoutError()
            except (zmq.error.ZMQError, asyncio.TimeoutError):
                return asyncio.create_task(
                    serv(port, datapath, url, buffer * 1024 * 1024, interval))
            await asyncio.sleep(timeout)


async def main(port, datapath, url, timeout=1, buffer=1024, interval=60):
    task = await watch(port=port,
                       datapath=datapath,
                       url=url,
                       timeout=timeout,
                       buffer=buffer,
                       interval=interval)
    await task


@click.command()
@click.option('--port', default=6789, help='Port of the server.')
@click.option('--datapath', default=datapath, help='Path of the data.')
@click.option('--url', default=None, help='URL of the database.')
@click.option('--timeout', default=1, help='Timeout of ping.')
@click.option('--buffer', default=1024, help='Buffer size (MB).')
@click.option('--interval',
              default=60,
              help='Interval of flush cache, in unit of second.')
def record(port, datapath, url, timeout, buffer, interval):
    asyncio.run(main(port, Path(datapath), url, timeout, buffer, interval))


if __name__ == "__main__":
    record()



from datetime import date, datetime, timezone
from typing import Sequence, Type, Union

from sqlalchemy.orm import Query, Session, aliased
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.session import Session
from waveforms.dicttree import foldDict

from .models import Comment, Record, Report, Sample, Tag


def tag(session: Session, tag_text: str) -> Tag:
    """Get a tag from the database or create a new if not exists."""
    try:
        return session.query(Tag).filter(Tag.text == tag_text).one()
    except NoResultFound:
        tag = Tag(text=tag_text)
        return tag


def tag_it(session: Session, tag_text: str, obj: Union[Sample, Record,
                                                       Report]) -> Tag:
    """Tag an object."""
    if obj.id is None:
        session.add(obj)
        obj.tags.append(tag(session, tag_text))
    else:
        session.query(type(obj)).filter(
            type(obj).id == obj.id).one().tags.append(tag(session, tag_text))
    session.commit()


def get_object_with_tags(session: Session,
                         cls: Union[Type[Comment], Type[Sample], Type[Record],
                                    Type[Report]], *tags: str) -> Query:
    """
    Query objects with the given tags.

    Parameters
    ----------
    session : :class:`sqlalchemy.orm.Session`
        The database session.
    cls : :class:`sqlalchemy.orm.Mapper`
        The object class.
    tags : str
        The tags.

    Returns
    -------
    :class:`sqlalchemy.orm.Query`
        The query.
    """
    if isinstance(session, Query):
        q = session
    else:
        q = session.query(cls)
    if not hasattr(cls, 'tags'):
        return []

    aliase = {tag: aliased(Tag) for tag in tags}

    for tag, a in aliase.items():
        q = q.join(a, cls.tags)
        if '*' in tag:
            q = q.filter(a.text.like(tag.replace('*', '%')))
        else:
            q = q.filter(a.text == tag)
    return q


def query_record(session: Session,
                 offset: int = 0,
                 limit: int = 10,
                 app: str | None = None,
                 tags: Sequence[str] = (),
                 before: datetime | date | None = None,
                 after: datetime | date | None = None):
    tz_offset = datetime.now(timezone.utc).astimezone().utcoffset()
    table = {'header': ['ID', 'App', 'tags', 'created time'], 'body': []}
    apps = sorted(
        set([
            n for n, *_ in get_object_with_tags(session.query(Record.app),
                                                Record, *tags).all()
        ]))
    apps = foldDict(dict([(app, None) for app in apps]))

    query = get_object_with_tags(session, Record, *tags)

    if app is not None:
        if app.endswith('*'):
            query = query.filter(Record.app.like(app[:-1] + '%'))
        else:
            query = query.filter(Record.app == app)
    if before is not None:
        if isinstance(before, date):
            before = datetime(before.year, before.month, before.day)
        query = query.filter(Record.ctime <= before - tz_offset)
    if after is not None:
        if isinstance(after, date):
            after = datetime(after.year, after.month, after.day)
        query = query.filter(Record.ctime >= after - tz_offset)
    total = query.count()
    for r in query.order_by(Record.ctime.desc()).limit(limit).offset(offset):
        tags = sorted([t.text for t in r.tags])
        ctime = r.ctime + tz_offset
        row = [r.id, r.app, tags, ctime]
        table['body'].append(row)

    return total, apps, table


def update_tags(session: Session,
                record_id: int,
                tags: Sequence[str],
                append: bool = False):
    record = session.get(Record, record_id)
    if record is None:
        return False
    if append:
        old = [t.text for t in record.tags]
        for t in old:
            if t not in tags:
                tags.append(t)
    record.tags = [tag(session, t) for t in tags]
    try:
        session.commit()
    except Exception:
        session.rollback()
        return False
    return True


def remove_tags(session: Session, record_id: int, tags: Sequence[str]):
    record = session.get(Record, record_id)
    if record is None:
        return False
    old = [t.text for t in record.tags]
    record.tags = [tag(session, t) for t in old if t not in tags]
    try:
        session.commit()
    except Exception:
        session.rollback()
        return False
    return True



import ast
import asyncio
import inspect
import itertools
import sys
import uuid
from graphlib import TopologicalSorter
from pathlib import Path
from types import MethodType
from typing import Any, Callable, Type

import dill
import numpy as np
import skopt
import zmq
from skopt.space import Categorical, Integer, Real
from tqdm.notebook import tqdm

from ..sys.rpc.zmq_socket import ZMQContextManager
from .expression import Env, Expression, Symbol
from .optimize import NgOptimizer
from .recorder import Record


async def call_function(func: Callable | Expression, variables: dict[str,
                                                                     Any]):
    if isinstance(func, Expression):
        env = Env()
        for name in func.symbols():
            if name in variables:
                if inspect.isawaitable(variables[name]):
                    variables[name] = await variables[name]
                env.variables[name] = variables[name]
            else:
                raise ValueError(f'{name} is not provided.')
        return func.eval(env)

    try:
        sig = inspect.signature(func)
    except:
        return func()
    args = []
    for name, param in sig.parameters.items():
        if param.kind == param.POSITIONAL_OR_KEYWORD:
            if name in variables:
                if inspect.isawaitable(variables[name]):
                    variables[name] = await variables[name]
                args.append(variables[name])
            elif param.default is not param.empty:
                args.append(param.default)
            else:
                raise ValueError(f'parameter {name} is not provided.')
        elif param.kind == param.VAR_POSITIONAL:
            raise ValueError('not support VAR_POSITIONAL')
        elif param.kind == param.VAR_KEYWORD:
            ret = func(**variables)
            if inspect.isawaitable(ret):
                ret = await ret
            return ret
    ret = func(*args)
    if inspect.isawaitable(ret):
        ret = await ret
    return ret


async def async_next(aiter):
    try:
        if hasattr(aiter, '__anext__'):
            return await aiter.__anext__()
        else:
            return next(aiter)
    except StopIteration:
        raise StopAsyncIteration from None


async def async_zip(*aiters):
    aiters = [
        ait.__aiter__() if hasattr(ait, '__aiter__') else iter(ait)
        for ait in aiters
    ]
    try:
        while True:
            # 使用 asyncio.gather 等待所有异步生成器返回下一个元素
            result = await asyncio.gather(*(async_next(ait) for ait in aiters))
            yield tuple(result)
    except StopAsyncIteration:
        # 当任一异步生成器耗尽时停止迭代
        return


def _get_depends(func: Callable):
    try:
        sig = inspect.signature(func)
    except:
        return []

    args = []
    for name, param in sig.parameters.items():
        if param.kind in [
                param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD,
                param.KEYWORD_ONLY
        ]:
            args.append(name)
        elif param.kind == param.VAR_KEYWORD:
            pass
        elif param.kind == param.VAR_POSITIONAL:
            raise ValueError('not support VAR_POSITIONAL')
    return args


def is_valid_identifier(s: str) -> bool:
    """
    Check if a string is a valid identifier.
    """
    try:
        ast.parse(f"f({s}=0)")
        return True
    except SyntaxError:
        return False


class OptimizeSpace():

    def __init__(self, optimizer: 'Optimizer', space):
        self.optimizer = optimizer
        self.space = space
        self.name = None

    def __len__(self):
        return self.optimizer.maxiter


class Optimizer():

    def __init__(self,
                 scanner: 'Scan',
                 name: str,
                 level: int,
                 method: str | Type = skopt.Optimizer,
                 maxiter: int = 1000,
                 minimize: bool = True,
                 **kwds):
        self.scanner = scanner
        self.method = method
        self.maxiter = maxiter
        self.dimensions = {}
        self.name = name
        self.level = level
        self.kwds = kwds
        self.minimize = minimize

    def create(self):
        return self.method(list(self.dimensions.values()), **self.kwds)

    def Categorical(self,
                    categories,
                    prior=None,
                    transform=None,
                    name=None) -> OptimizeSpace:
        return OptimizeSpace(self,
                             Categorical(categories, prior, transform, name))

    def Integer(self,
                low,
                high,
                prior="uniform",
                base=10,
                transform=None,
                name=None,
                dtype=np.int64) -> OptimizeSpace:
        return OptimizeSpace(
            self, Integer(low, high, prior, base, transform, name, dtype))

    def Real(self,
             low,
             high,
             prior="uniform",
             base=10,
             transform=None,
             name=None,
             dtype=float) -> OptimizeSpace:
        return OptimizeSpace(
            self, Real(low, high, prior, base, transform, name, dtype))

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        del state['scanner']
        return state

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)
        self.scanner = None


class Promise():
    __slots__ = ['task', 'key', 'attr']

    def __init__(self, task, key=None, attr=None):
        self.task = task
        self.key = key
        self.attr = attr

    def __await__(self):

        async def _getitem(task, key):
            return (await task)[key]

        async def _getattr(task, attr):
            return getattr(await task, attr)

        if self.key is not None:
            return _getitem(self.task, self.key).__await__()
        elif self.attr is not None:
            return _getattr(self.task, self.attr).__await__()
        else:
            return self.task.__await__()

    def __getitem__(self, key):
        return Promise(self.task, key, None)

    def __getattr__(self, attr):
        return Promise(self.task, None, attr)


class Scan():

    def __init__(self,
                 app: str = 'task',
                 tags: tuple[str] = (),
                 database: str | Path | None = 'tcp://127.0.0.1:6789'):
        self.id = f"{app}({str(uuid.uuid1())})"
        self.record = None
        self.namespace = {}
        self.description = {
            'app': app,
            'tags': tags,
            'loops': {},
            'consts': {},
            'functions': {},
            'optimizers': {},
            'actions': {},
            'dependents': {},
            'order': {},
            'filters': {},
            'total': {},
            'compiled': False,
        }
        self._current_level = 0
        self.variables = {}
        self._task = None
        self.sock = None
        self.database = database
        self._variables = {}
        self._sem = asyncio.Semaphore(100)
        self._bar = {}

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        del state['record']
        del state['sock']
        del state['_task']
        del state['_sem']
        return state

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)
        self.record = None
        self.sock = None
        self._task = None
        self._sem = asyncio.Semaphore(100)
        for opt in self.description['optimizers'].values():
            opt.scanner = self

    @property
    def current_level(self):
        return self._current_level

    async def emit(self, current_level, step, position, variables: dict[str,
                                                                        Any]):
        for key, value in list(variables.items()):
            if inspect.isawaitable(value):
                variables[key] = await value
        if self.sock is not None:
            await self.sock.send_pyobj({
                'task': self.id,
                'method': 'record_append',
                'record_id': self.record.id,
                'level': current_level,
                'step': step,
                'position': position,
                'variables': variables
            })
        else:
            self.record.append(current_level, step, position, variables)

    async def _filter(self, variables: dict[str, Any], level: int = 0):
        try:
            return all([
                await call_function(fun, variables) for fun in itertools.chain(
                    self.description['filters'].get(level, []),
                    self.description['filters'].get(-1, []))
            ])
        except:
            return True

    async def create_record(self):
        import __main__
        from IPython import get_ipython

        ipy = get_ipython()
        if ipy is not None:
            scripts = ('ipython', ipy.user_ns['In'])
        else:
            try:
                scripts = ('shell',
                           [sys.executable, __main__.__file__, *sys.argv[1:]])
            except:
                scripts = ('', [])

        if self.sock is not None:
            await self.sock.send_pyobj({
                'task':
                self.id,
                'method':
                'record_create',
                'description':
                dill.dumps(self.description),
                # 'env':
                # dill.dumps(__main__.__dict__),
                'scripts':
                scripts,
                'tags': []
            })

            record_id = await self.sock.recv_pyobj()
            return Record(record_id, self.database, self.description)
        return Record(None, self.database, self.description)

    def get(self, name: str):
        if name in self.description['consts']:
            return self.description['consts'][name]
        elif name in self.namespace:
            return self.namespace.get(name)
        else:
            return Symbol(name)

    def _add_loop_var(self, name: str, level: int, range):
        if level not in self.description['loops']:
            self.description['loops'][level] = []
        self.description['loops'][level].append((name, range))

    def add_depends(self, name: str, depends: list[str]):
        if isinstance(depends, str):
            depends = [depends]
        if name not in self.description['dependents']:
            self.description['dependents'][name] = set()
        self.description['dependents'][name].update(depends)

    def add_filter(self, func, level):
        """
        Add a filter function to the scan.

        Args:
            func: A callable object or an instance of Expression.
            level: The level of the scan to add the filter. -1 means any level.
        """
        if level not in self.description['filters']:
            self.description['filters'][level] = []
        self.description['filters'][level].append(func)

    def set(self, name: str, value):
        if isinstance(value, Expression):
            self.add_depends(name, value.symbols())
            self.description['functions'][name] = value
        elif callable(value):
            self.add_depends(name, _get_depends(value))
            self.description['functions'][name] = value
        else:
            self.description['consts'][name] = value

    def search(self, name: str, range, level: int | None = None):
        if level is not None:
            assert level >= 0, 'level must be greater than or equal to 0.'
        if isinstance(range, OptimizeSpace):
            range.name = name
            range.optimizer.dimensions[name] = range.space
            self._add_loop_var(name, range.optimizer.level, range)
            self.add_depends(range.optimizer.name, [name])
        else:
            if level is None:
                raise ValueError('level must be provided.')
            self._add_loop_var(name, level, range)
            if isinstance(range, Expression) or callable(range):
                self.add_depends(name, range.symbols())

    def minimize(self,
                 name: str,
                 level: int,
                 method=NgOptimizer,
                 maxiter=100,
                 **kwds) -> Optimizer:
        assert level >= 0, 'level must be greater than or equal to 0.'
        opt = Optimizer(self,
                        name,
                        level,
                        method,
                        maxiter,
                        minimize=True,
                        **kwds)
        self.description['optimizers'][name] = opt
        return opt

    def maximize(self,
                 name: str,
                 level: int,
                 method=NgOptimizer,
                 maxiter=100,
                 **kwds) -> Optimizer:
        assert level >= 0, 'level must be greater than or equal to 0.'
        opt = Optimizer(self,
                        name,
                        level,
                        method,
                        maxiter,
                        minimize=False,
                        **kwds)
        self.description['optimizers'][name] = opt
        return opt

    async def _run(self):
        self.assymbly()
        self.variables = self.description['consts'].copy()
        for level, total in self.description['total'].items():
            if total == np.inf:
                total = None
            self._bar[level] = tqdm(total=total)
        for group in self.description['order'].get(-1, []):
            for name in group:
                if name in self.description['functions']:
                    self.variables[name] = await call_function(
                        self.description['functions'][name], self.variables)
        if isinstance(self.database,
                      str) and self.database.startswith("tcp://"):
            async with ZMQContextManager(zmq.DEALER,
                                         connect=self.database) as socket:
                self.sock = socket
                self.record = await self.create_record()
                await self.work()
        else:
            self.record = await self.create_record()
            await self.work()
        for level, bar in self._bar.items():
            bar.close()
        return self.variables

    async def done(self):
        if self._task is not None:
            try:
                await self._task
            except asyncio.CancelledError:
                pass

    def start(self):
        import asyncio
        self._task = asyncio.create_task(self._run())

    def cancel(self):
        if self._task is not None:
            self._task.cancel()

    def assymbly(self):
        if self.description['compiled']:
            return

        mapping = {
            label: level
            for level, label in enumerate(
                sorted(
                    set(self.description['loops'].keys())
                    | set(self.description['actions'].keys()) - {-1}))
        }

        if -1 in self.description['actions']:
            mapping[-1] = max(mapping.values()) + 1

        self.description['loops'] = dict(
            sorted([(mapping[k], v)
                    for k, v in self.description['loops'].items()]))
        self.description['actions'] = {
            mapping[k]: v
            for k, v in self.description['actions'].items()
        }

        for level, loops in self.description['loops'].items():
            self.description['total'][level] = np.inf
            for name, space in loops:
                try:
                    self.description['total'][level] = min(
                        self.description['total'][level], len(space))
                except:
                    pass

        dependents = self.description['dependents'].copy()

        for level in range(len(mapping)):
            range_list = self.description['loops'].get(level, [])
            if level > 0:
                if f'#__loop_{level}' not in self.description['dependents']:
                    dependents[f'#__loop_{level}'] = []
                dependents[f'#__loop_{level}'].append(f'#__loop_{level-1}')
            for name, _ in range_list:
                if name not in self.description['dependents']:
                    dependents[name] = []
                dependents[name].append(f'#__loop_{level}')

        def _get_all_depends(key, graph):
            ret = set()
            if key not in graph:
                return ret

            for e in graph[key]:
                ret.update(_get_all_depends(e, graph))
            ret.update(graph[key])
            return ret

        full_depends = {}
        for key in dependents:
            full_depends[key] = _get_all_depends(key, dependents)

        levels = {}
        passed = set()
        all_keys = set()
        for level in reversed(self.description['loops'].keys()):
            tag = f'#__loop_{level}'
            for key, deps in full_depends.items():
                all_keys.update(deps)
                all_keys.add(key)
                if key.startswith('#__loop_'):
                    continue
                if tag in deps:
                    if level not in levels:
                        levels[level] = set()
                    if key not in passed:
                        passed.add(key)
                        levels[level].add(key)
        levels[-1] = {
            key
            for key in all_keys - passed if not key.startswith('#__loop_')
        }

        order = []
        ts = TopologicalSorter(dependents)
        ts.prepare()
        while ts.is_active():
            ready = ts.get_ready()
            order.append(ready)
            for k in ready:
                ts.done(k)

        self.description['order'] = {}

        for level in sorted(levels):
            keys = set(levels[level])
            self.description['order'][level] = []
            for ready in order:
                ready = list(keys & set(ready))
                if ready:
                    self.description['order'][level].append(ready)
                    keys -= set(ready)

        self.description['compiled'] = True

    async def _iter_level(self, level, variables):
        iters = {}
        env = Env()
        env.variables = variables
        opts = {}

        for name, iter in self.description['loops'][level]:
            if isinstance(iter, OptimizeSpace):
                if iter.optimizer.name not in opts:
                    opts[iter.optimizer.name] = iter.optimizer.create()
            elif isinstance(iter, Expression):
                iters[name] = iter.eval(env)
            elif callable(iter):
                iters[name] = await call_function(iter, variables)
            else:
                iters[name] = iter

        maxiter = 0xffffffff
        for name, opt in opts.items():
            opt_cfg = self.description['optimizers'][name]
            maxiter = min(maxiter, opt_cfg.maxiter)

        async for args in async_zip(*iters.values(), range(maxiter)):
            variables.update(dict(zip(iters.keys(), args[:-1])))
            for name, opt in opts.items():
                args = opt.ask()
                opt_cfg = self.description['optimizers'][name]
                variables.update({
                    n: v
                    for n, v in zip(opt_cfg.dimensions.keys(), args)
                })

            for group in self.description['order'].get(level, []):
                for name in group:
                    if name in self.description['functions']:
                        variables[name] = await call_function(
                            self.description['functions'][name], variables)

            yield variables

            for name, opt in opts.items():
                opt_cfg = self.description['optimizers'][name]
                args = [variables[n] for n in opt_cfg.dimensions.keys()]
                if name not in variables:
                    raise ValueError(f'{name} not in variables.')
                fun = variables[name]
                if inspect.isawaitable(fun):
                    fun = await fun
                if opt_cfg.minimize:
                    opt.tell(args, fun)
                else:
                    opt.tell(args, -fun)

        for name, opt in opts.items():
            opt_cfg = self.description['optimizers'][name]
            result = opt.get_result()
            variables.update({
                n: v
                for n, v in zip(opt_cfg.dimensions.keys(), result.x)
            })
            variables[name] = result.fun
        if opts:
            yield variables

    async def iter(self, **kwds):
        if self.current_level >= len(self.description['loops']):
            return
        step = 0
        position = 0
        task = None
        if self.current_level in self._bar:
            self._bar[self.current_level].reset()
        async for variables in self._iter_level(self.current_level,
                                                self.variables):
            self._current_level += 1
            if await self._filter(variables, self.current_level - 1):
                yield variables
                task = asyncio.create_task(
                    self.emit(self.current_level - 1, step, position,
                              variables.copy()))
                step += 1
            position += 1
            self._current_level -= 1
            if self.current_level in self._bar:
                self._bar[self.current_level].update(1)
        if task is not None:
            await task
        if self.current_level == 0:
            await self.emit(self.current_level - 1, 0, 0, {})

    async def work(self, **kwds):
        if self.current_level in self.description['actions']:
            action = self.description['actions'][self.current_level]
            coro = action(self, **kwds)
            if inspect.isawaitable(coro):
                await coro
        else:
            async for variables in self.iter(**kwds):
                await self.do_something(**kwds)

    async def do_something(self, **kwds):
        await self.work(**kwds)

    def mount(self, action: Callable, level: int):
        """
        Mount a action to the scan.

        Args:
            action: A callable object.
            level: The level of the scan to mount the action.
        """
        self.description['actions'][level] = action

    async def promise(self, awaitable):
        """
        Promise to calculate asynchronous function and return the result in future.

        Args:
            awaitable: An awaitable object.

        Returns:
            Promise: A promise object.
        """
        async with self._sem:
            return Promise(asyncio.create_task(self._await(awaitable)))

    async def _await(self, awaitable):
        async with self._sem:
            return await awaitable
